@@ -63,7 +63,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   /** Returns the number of bytes currently in this buffer. */
-  public long size() {
+  public final long size() {
     return size;
   }
 
@@ -138,7 +138,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   /** Copy the contents of this to {@code out}. */
-  public Buffer copyTo(OutputStream out) throws IOException {
+  public final Buffer copyTo(OutputStream out) throws IOException {
     return copyTo(out, 0, size);
   }
 
@@ -146,7 +146,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
    * Copy {@code byteCount} bytes from this, starting at {@code offset}, to
    * {@code out}.
    */
-  public Buffer copyTo(OutputStream out, long offset, long byteCount) throws IOException {
+  public final Buffer copyTo(OutputStream out, long offset, long byteCount) throws IOException {
     if (out == null) throw new IllegalArgumentException("out == null");
     checkOffsetAndCount(size, offset, byteCount);
     if (byteCount == 0) return this;
@@ -170,7 +170,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   /** Copy {@code byteCount} bytes from this, starting at {@code offset}, to {@code out}. */
-  public Buffer copyTo(Buffer out, long offset, long byteCount) {
+  public final Buffer copyTo(Buffer out, long offset, long byteCount) {
     if (out == null) throw new IllegalArgumentException("out == null");
     checkOffsetAndCount(size, offset, byteCount);
     if (byteCount == 0) return this;
@@ -201,12 +201,12 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   /** Write the contents of this to {@code out}. */
-  public Buffer writeTo(OutputStream out) throws IOException {
+  public final Buffer writeTo(OutputStream out) throws IOException {
     return writeTo(out, size);
   }
 
   /** Write {@code byteCount} bytes from this to {@code out}. */
-  public Buffer writeTo(OutputStream out, long byteCount) throws IOException {
+  public final Buffer writeTo(OutputStream out, long byteCount) throws IOException {
     if (out == null) throw new IllegalArgumentException("out == null");
     checkOffsetAndCount(size, 0, byteCount);
 
@@ -230,13 +230,13 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   /** Read and exhaust bytes from {@code in} to this. */
-  public Buffer readFrom(InputStream in) throws IOException {
+  public final Buffer readFrom(InputStream in) throws IOException {
     readFrom(in, Long.MAX_VALUE, true);
     return this;
   }
 
   /** Read {@code byteCount} bytes from {@code in} to this. */
-  public Buffer readFrom(InputStream in, long byteCount) throws IOException {
+  public final Buffer readFrom(InputStream in, long byteCount) throws IOException {
     if (byteCount < 0) throw new IllegalArgumentException("byteCount < 0: " + byteCount);
     readFrom(in, byteCount, false);
     return this;
@@ -263,7 +263,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
    * number of bytes that can be flushed immediately to an underlying sink
    * without harming throughput.
    */
-  public long completeSegmentByteCount() {
+  public final long completeSegmentByteCount() {
     long result = size;
     if (result == 0) return 0;
 
@@ -298,7 +298,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   /** Returns the byte at {@code pos}. */
-  public byte getByte(long pos) {
+  public final byte getByte(long pos) {
     checkOffsetAndCount(size, pos, 1);
     if (size - pos > pos) {
       for (Segment s = head; true; s = s.next) {
@@ -545,40 +545,122 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   @Override public int select(Options options) {
-    Segment s = head;
-    if (s == null) return options.indexOf(ByteString.EMPTY);
+    int index = selectPrefix(options, false);
+    if (index == -1) return -1;
 
-    ByteString[] byteStrings = options.byteStrings;
-    for (int i = 0, listSize = byteStrings.length; i < listSize; i++) {
-      ByteString b = byteStrings[i];
-      if (size >= b.size() && rangeEquals(s, s.pos, b, 0, b.size())) {
-        try {
-          skip(b.size());
-          return i;
-        } catch (EOFException e) {
-          throw new AssertionError(e);
-        }
-      }
+    // If the prefix match actually matched a full byte string, consume it and return it.
+    int selectedSize = options.byteStrings[index].size();
+    try {
+      skip(selectedSize);
+    } catch (EOFException e) {
+      throw new AssertionError();
     }
-    return -1;
+    return index;
   }
 
   /**
-   * Returns the index of a value in {@code options} that is either the prefix of this buffer, or
-   * that this buffer is a prefix of. Unlike {@link #select} this never consumes the value, even
-   * if it is found in full.
+   * Returns the index of a value in options that is a prefix of this buffer. Returns -1 if no value
+   * is found. This method does two simultaneous iterations: it iterates the trie and it iterates
+   * this buffer. It returns when it reaches a result in the trie, when it mismatches in the trie,
+   * and when the buffer is exhausted.
+   *
+   * @param selectTruncated true to return -2 if a possible result is present but truncated. For
+   *     example, this will return -2 if the buffer contains [ab] and the options are [abc, abd].
+   *     Note that this is made complicated by the fact that options are listed in preference order,
+   *     and one option may be a prefix of another. For example, this returns -2 if the buffer
+   *     contains [ab] and the options are [abc, a].
    */
-  int selectPrefix(Options options) {
-    Segment s = head;
-    ByteString[] byteStrings = options.byteStrings;
-    for (int i = 0, listSize = byteStrings.length; i < listSize; i++) {
-      ByteString b = byteStrings[i];
-      int bytesLimit = (int) Math.min(size, b.size());
-      if (bytesLimit == 0 || rangeEquals(s, s.pos, b, 0, bytesLimit)) {
-        return i;
-      }
+  int selectPrefix(Options options, boolean selectTruncated) {
+    Segment head = this.head;
+    if (head == null) {
+      if (selectTruncated) return -2; // A result is present but truncated.
+      return options.indexOf(ByteString.EMPTY);
     }
-    return -1;
+
+    Segment s = head;
+    byte[] data = head.data;
+    int pos = head.pos;
+    int limit = head.limit;
+
+    int[] trie = options.trie;
+    int triePos = 0;
+
+    int prefixIndex = -1;
+
+    navigateTrie:
+    while (true) {
+      int scanOrSelect = trie[triePos++];
+
+      int possiblePrefixIndex = trie[triePos++];
+      if (possiblePrefixIndex != -1) {
+        prefixIndex = possiblePrefixIndex;
+      }
+
+      int nextStep;
+
+      if (s == null) {
+        break;
+      } else if (scanOrSelect < 0) {
+        // Scan: take multiple bytes from the buffer and the trie, looking for any mismatch.
+        int scanByteCount = -1 * scanOrSelect;
+        int trieLimit = triePos + scanByteCount;
+        while (true) {
+          int b = data[pos++] & 0xff;
+          if (b != trie[triePos++]) return prefixIndex; // Fail 'cause we found a mismatch.
+          boolean scanComplete = (triePos == trieLimit);
+
+          // Advance to the next buffer segment if this one is exhausted.
+          if (pos == limit) {
+            s = s.next;
+            pos = s.pos;
+            data = s.data;
+            limit = s.limit;
+            if (s == head) {
+              if (!scanComplete) break navigateTrie; // We were exhausted before the scan completed.
+              s = null; // We were exhausted at the end of the scan.
+            }
+          }
+
+          if (scanComplete) {
+            nextStep = trie[triePos];
+            break;
+          }
+        }
+      } else {
+        // Select: take one byte from the buffer and find a match in the trie.
+        int selectChoiceCount = scanOrSelect;
+        int b = data[pos++] & 0xff;
+        int selectLimit = triePos + selectChoiceCount;
+        while (true) {
+          if (triePos == selectLimit) return prefixIndex; // Fail 'cause we didn't find a match.
+
+          if (b == trie[triePos]) {
+            nextStep = trie[triePos + selectChoiceCount];
+            break;
+          }
+
+          triePos++;
+        }
+
+        // Advance to the next buffer segment if this one is exhausted.
+        if (pos == limit) {
+          s = s.next;
+          pos = s.pos;
+          data = s.data;
+          limit = s.limit;
+          if (s == head) {
+            s = null; // No more segments! The next trie node will be our last.
+          }
+        }
+      }
+
+      if (nextStep >= 0) return nextStep; // Found a matching option.
+      triePos = -nextStep; // Found another node to continue the search.
+    }
+
+    // We break out of the loop above when we've exhausted the buffer without exhausting the trie.
+    if (selectTruncated) return -2; // The buffer is a prefix of at least one option.
+    return prefixIndex; // Return any matches we encountered while searching for a deeper match.
   }
 
   @Override public void readFully(Buffer sink, long byteCount) throws EOFException {
@@ -835,7 +917,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
    * Discards all bytes in this buffer. Calling this method when you're done
    * with a buffer will return its segments to the pool.
    */
-  public void clear() {
+  public final void clear() {
     try {
       skip(size);
     } catch (EOFException e) {
@@ -1598,22 +1680,22 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   /** Returns the 128-bit MD5 hash of this buffer. */
-  public ByteString md5() {
+  public final ByteString md5() {
     return digest("MD5");
   }
 
   /** Returns the 160-bit SHA-1 hash of this buffer. */
-  public ByteString sha1() {
+  public final ByteString sha1() {
     return digest("SHA-1");
   }
 
   /** Returns the 256-bit SHA-256 hash of this buffer. */
-  public ByteString sha256() {
+  public final ByteString sha256() {
     return digest("SHA-256");
   }
 
   /** Returns the 512-bit SHA-512 hash of this buffer. */
-  public ByteString sha512() {
+  public final ByteString sha512() {
       return digest("SHA-512");
   }
 
@@ -1633,17 +1715,17 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   /** Returns the 160-bit SHA-1 HMAC of this buffer. */
-  public ByteString hmacSha1(ByteString key) {
+  public final ByteString hmacSha1(ByteString key) {
     return hmac("HmacSHA1", key);
   }
 
   /** Returns the 256-bit SHA-256 HMAC of this buffer. */
-  public ByteString hmacSha256(ByteString key) {
+  public final ByteString hmacSha256(ByteString key) {
     return hmac("HmacSHA256", key);
   }
 
   /** Returns the 512-bit SHA-512 HMAC of this buffer. */
-  public ByteString hmacSha512(ByteString key) {
+  public final ByteString hmacSha512(ByteString key) {
       return hmac("HmacSHA512", key);
   }
 
@@ -1734,7 +1816,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   }
 
   /** Returns an immutable copy of this buffer as a byte string. */
-  public ByteString snapshot() {
+  public final ByteString snapshot() {
     if (size > Integer.MAX_VALUE) {
       throw new IllegalArgumentException("size > Integer.MAX_VALUE: " + size);
     }
@@ -1744,16 +1826,16 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
   /**
    * Returns an immutable copy of the first {@code byteCount} bytes of this buffer as a byte string.
    */
-  public ByteString snapshot(int byteCount) {
+  public final ByteString snapshot(int byteCount) {
     if (byteCount == 0) return ByteString.EMPTY;
     return new SegmentedByteString(this, byteCount);
   }
 
-  public UnsafeCursor readUnsafe() {
+  public final UnsafeCursor readUnsafe() {
     return readUnsafe(new UnsafeCursor());
   }
 
-  public UnsafeCursor readUnsafe(UnsafeCursor unsafeCursor) {
+  public final UnsafeCursor readUnsafe(UnsafeCursor unsafeCursor) {
     if (unsafeCursor.buffer != null) {
       throw new IllegalStateException("already attached to a buffer");
     }
@@ -1763,11 +1845,11 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
     return unsafeCursor;
   }
 
-  public UnsafeCursor readAndWriteUnsafe() {
+  public final UnsafeCursor readAndWriteUnsafe() {
     return readAndWriteUnsafe(new UnsafeCursor());
   }
 
-  public UnsafeCursor readAndWriteUnsafe(UnsafeCursor unsafeCursor) {
+  public final UnsafeCursor readAndWriteUnsafe(UnsafeCursor unsafeCursor) {
     if (unsafeCursor.buffer != null) {
       throw new IllegalStateException("already attached to a buffer");
     }
@@ -1992,7 +2074,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
      * size of the readable range (at least 1), or -1 if we have reached the end of the buffer and
      * there are no more bytes to read.
      */
-    public int next() {
+    public final int next() {
       if (offset == buffer.size) throw new IllegalStateException();
       if (offset == -1L) return seek(0L);
       return seek(offset + (end - start));
@@ -2003,7 +2085,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
      * Returns the number of bytes readable in {@code data} (at least 1), or -1 if there are no data
      * to read.
      */
-    public int seek(long offset) {
+    public final int seek(long offset) {
       if (offset < -1 || offset > buffer.size) {
         throw new ArrayIndexOutOfBoundsException(
             String.format("offset=%s > size=%s", offset, buffer.size));
@@ -2092,7 +2174,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
      *
      * @return the previous size of the buffer.
      */
-    public long resizeBuffer(long newSize) {
+    public final long resizeBuffer(long newSize) {
       if (buffer == null) {
         throw new IllegalStateException("not attached to a buffer");
       }
@@ -2173,7 +2255,7 @@ public final class Buffer implements BufferedSource, BufferedSink, Cloneable, By
      *     than the capacity size of a single segment (8 KiB).
      * @return the number of bytes expanded by. Not less than {@code minByteCount}.
      */
-    public long expandBuffer(int minByteCount) {
+    public final long expandBuffer(int minByteCount) {
       if (minByteCount <= 0) {
         throw new IllegalArgumentException("minByteCount <= 0: " + minByteCount);
       }
